# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Code that draws the box around a Pawn or Spot when it's selected"""

from collections import defaultdict
from functools import partial
from operator import itemgetter
from time import monotonic

import numpy as np
from kivy.core.image import Image
from kivy.graphics.fbo import Fbo
from kivy.properties import ObjectProperty, BooleanProperty, ListProperty
from kivy.graphics import (
	InstructionGroup,
	Translate,
	PopMatrix,
	PushMatrix,
	Color,
	Line,
	Rectangle,
)
from kivy.resources import resource_find
from kivy.uix.layout import Layout
from kivy.clock import Clock, mainthread, triggered
from kivy.uix.widget import Widget
from kivy.logger import Logger

from ELiDE.imagestackproxy import ImageStackProxy


def trigger(func):
	return triggered()(func)


class TextureStackPlane(Widget):
	data = ListProperty()
	selected = ObjectProperty(allownone=True)
	color_selected = ListProperty([0.0, 1.0, 1.0, 1.0])
	default_image_path = "atlas://rltiles/base.atlas/unseen"

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self._keys = []
		self._left_xs = np.array([])
		self._right_xs = np.array([])
		self._top_ys = np.array([])
		self._bot_ys = np.array([])
		self._instructions = {}
		self._stack_index = {}
		self._trigger_redraw = Clock.create_trigger(self.redraw)
		self._redraw_bind_uid = self.fbind("data", self._trigger_redraw)

	def on_parent(self, *_):
		if not self.canvas:
			Clock.schedule_once(self.on_parent, 0)
			return
		with self.canvas:
			self._fbo = Fbo(size=self.size)
			self._translate = Translate(x=self.x, y=self.y)
			self._rectangle = Rectangle(
				size=self.size, texture=self._fbo.texture
			)
		self.bind(pos=self._trigger_redraw, size=self._trigger_redraw)
		self._trigger_redraw()

	def on_pos(self, *_):
		if not hasattr(self, "_translate"):
			return
		self._translate.x, self._translate.y = self.pos
		self.canvas.ask_update()

	def on_size(self, *_):
		if not hasattr(self, "_rectangle") or not hasattr(self, "_fbo"):
			return
		self._rectangle.size = self._fbo.size = self.size
		self.redraw()

	@mainthread
	def _add_datum_upd_fbo(self, **datum):
		name = datum["name"]
		texs = datum["textures"]
		x = datum["x"]
		y = datum["y"]
		fbo = self._fbo
		with fbo:
			instructions = self._instructions
			rects = []
			wide = datum.get("width", 0)
			tall = datum.get("height", 0)
			for tex in texs:
				if isinstance(tex, str):
					tex = Image.load(resource_find(tex)).texture
					w, h = tex.size
					if "width" not in datum and w > wide:
						wide = w
					if "height" not in datum and h > tall:
						tall = h
				rects.append(
					Rectangle(texture=tex, pos=(x, y), size=(wide, tall))
				)
			instructions[name] = {
				"rectangles": rects,
				"group": InstructionGroup(),
			}
			grp = instructions[name]["group"]
			for rect in rects:
				grp.add(rect)
			fbo.add(instructions[name]["group"])

	def add_datum(self, datum):
		name = datum["name"]
		if "pos" in datum:
			x, y = datum.pop("pos")
		else:
			x = datum["x"]
			y = datum["y"]
		if isinstance(x, float):
			x *= self.width
		if isinstance(y, float):
			y *= self.height
		left_xs = list(self._left_xs)
		right_xs = list(self._right_xs)
		top_ys = list(self._top_ys)
		bot_ys = list(self._bot_ys)
		left_xs.append(x)
		bot_ys.append(y)
		wide = datum.get("width", 0)
		tall = datum.get("height", 0)
		top_ys.append(y + tall)
		right_xs.append(x + wide)
		self._left_xs = np.array(left_xs)
		self._bot_ys = np.array(bot_ys)
		self._top_ys = np.array(top_ys)
		self._right_xs = np.array(right_xs)
		self._stack_index[name] = len(self._keys)
		self._keys.append(name)
		self.unbind_uid("data", self._redraw_bind_uid)
		self.data.append(datum)
		self._redraw_bind_uid = self.fbind("data", self._trigger_redraw)
		self._add_datum_upd_fbo(**datum)

	@mainthread
	def _remove_upd_fbo(self, name):
		if name not in self._instructions:
			return
		grp = self._instructions[name]["group"]
		grp.clear()
		fbo = self._fbo
		fbo.bind()
		fbo.remove(grp)
		fbo.ask_update()
		fbo.clear_buffer()
		fbo.release()
		self._rectangle.texture = fbo.texture
		del self._instructions[name]

	def remove(self, name_or_idx):
		def delarr(arr, i):
			if i == 0:
				return arr[1:]
			elif i == len(arr) - 1:
				return arr[:-1]
			else:
				return np.concatenate((arr[:i], arr[i + 1 :]))

		if name_or_idx in self._keys:
			idx = self._keys.index(name_or_idx)
			name = name_or_idx
		else:
			idx = name_or_idx
			name = self._keys[idx]
		stack_index = self._stack_index
		del stack_index[name]
		del self._keys[idx]
		for key in self._keys[idx:]:
			stack_index[key] -= 1
		self._left_xs = delarr(self._left_xs, idx)
		self._bot_ys = delarr(self._bot_ys, idx)
		self._top_ys = delarr(self._top_ys, idx)
		self._right_xs = delarr(self._right_xs, idx)
		self.unbind_uid("data", self._redraw_bind_uid)
		del self.data[idx]
		self._redraw_bind_uid = self.fbind("data", self._trigger_redraw)
		self._remove_upd_fbo(name)

	@mainthread
	def _redraw_upd_fbo(self, changed_instructions):
		fbo = self._fbo
		for insts in changed_instructions:
			group = insts["group"]
			group.clear()
			for rect in insts["rectangles"]:
				group.add(rect)
			if "color0" in insts:
				group.add(insts["color0"])
				group.add(insts["line"])
				group.add(insts["color1"])
			if group not in fbo.children:
				fbo.add(group)
		self._rectangle.texture = fbo.texture

	@mainthread
	def _redraw_remove_fbo(self, removed_instructions):
		fbo = self._fbo
		for insts in removed_instructions:
			fbo.remove(insts["group"])

	def redraw(self, *_):
		def get_rects(datum):
			width = datum.get("width", 0)
			height = datum.get("height", 0)
			if isinstance(datum["x"], float):
				x = datum["x"] * self_width
			else:
				if not isinstance(datum["x"], int):
					raise TypeError("need int or float for pos")
				x = datum["x"]
			if isinstance(datum["y"], float):
				y = datum["y"] * self_height
			else:
				if not isinstance(datum["y"], int):
					raise TypeError("need int or float for pos")
				y = datum["y"]
			rects = []
			for texture in datum["textures"]:
				if isinstance(texture, str):
					try:
						texture = Image.load(resource_find(texture)).texture
					except Exception:
						texture = Image.load(self.default_image_path).texture
				w, h = texture.size
				if "width" in datum:
					w = width
				elif w > width:
					width = w
				if "height" in datum:
					h = height
				elif h > height:
					height = h
				assert w > 0 and h > 0
				rects.append(
					Rectangle(pos=(x, y), size=(w, h), texture=texture)
				)
			return rects

		def get_lines_and_colors(datum) -> dict:
			width = datum.get("width", 0)
			height = datum.get("height", 0)
			if isinstance(datum["x"], float):
				x = datum["x"] * self_width
			else:
				if not isinstance(datum["x"], int):
					raise TypeError("need int or float for pos")
				x = datum["x"]
			if isinstance(datum["y"], float):
				y = datum["y"] * self_height
			else:
				if not isinstance(datum["y"], int):
					raise TypeError("need int or float for pos")
				y = datum["y"]
			right = x + width
			top = y + height
			instructions = {}
			colr = Color(rgba=color_selected)
			instructions["color0"] = colr
			line = Line(points=[x, y, right, y, right, top, x, top, x, y])
			instructions["line"] = line
			coler = Color(rgba=[1, 1, 1, 1])
			instructions["color1"] = coler
			return instructions

		if not hasattr(self, "_rectangle"):
			self._trigger_redraw()
			return
		Logger.debug("TextureStackPlane: redrawing")
		start_ts = monotonic()
		instructions = self._instructions
		stack_index = self._stack_index
		keys = list(self._keys)
		left_xs = list(self._left_xs)
		right_xs = list(self._right_xs)
		top_ys = list(self._top_ys)
		bot_ys = list(self._bot_ys)
		self_width = self.width
		self_height = self.height
		selected = self.selected
		color_selected = self.color_selected
		todo = []
		observed = set()
		for datum in self.data:
			name = datum["name"]
			observed.add(name)
			texs = datum["textures"]
			if isinstance(datum["x"], float):
				x = datum["x"] * self_width
			else:
				if not isinstance(datum["x"], int):
					raise TypeError("need int or float for pos")
				x = datum["x"]
			if isinstance(datum["y"], float):
				y = datum["y"] * self_height
			else:
				if not isinstance(datum["y"], int):
					raise TypeError("need int or float for pos")
				y = datum["y"]
			if name in stack_index:
				rects = get_rects(datum)
				if name == selected:
					insts = get_lines_and_colors(datum)
				else:
					insts = {}
				insts["rectangles"] = rects
				if name in instructions:
					insts["group"] = instructions[name]["group"]
				else:
					insts["group"] = InstructionGroup()
				todo.append(insts)
				instructions[name] = insts
				width = datum.get("width", 0)
				height = datum.get("height", 0)
				for texture, rect in zip(texs, rects):
					if isinstance(texture, str):
						try:
							texture = Image.load(
								resource_find(texture)
							).texture
						except Exception:
							texture = Image.load(
								self.default_image_path
							).texture
					w, h = texture.size
					if "width" in datum:
						w = width
					elif w > width:
						width = w
					if "height" in datum:
						h = height
					elif h > height:
						height = h
					rect.texture = texture
					assert w > 0 and h > 0
					rect.size = (w, h)
				idx = stack_index[name]
				right = x + width
				left_xs[idx] = x
				right_xs[idx] = right
				bot_ys[idx] = y
				top = y + height
				top_ys[idx] = top
			else:
				width = datum.get("width", 0)
				height = datum.get("height", 0)
				stack_index[name] = len(keys)
				keys.append(name)
				rects = get_rects(datum)
				grp = InstructionGroup()
				if "width" not in datum or "height" not in datum:
					width, height = rects[0].size
				right = x + width
				left_xs.append(x)
				right_xs.append(right)
				bot_ys.append(y)
				top = y + height
				top_ys.append(top)
				instructions[name] = insts = {
					"rectangles": rects,
					"group": grp,
				}
				if name == selected:
					insts.update(get_lines_and_colors())
				todo.append(insts)
		unobserved = instructions.keys() - observed
		get_rid = []
		for gone in unobserved:
			get_rid.append(instructions.pop(gone))
		self._left_xs = np.array(left_xs)
		self._right_xs = np.array(right_xs)
		self._top_ys = np.array(top_ys)
		self._bot_ys = np.array(bot_ys)
		self._keys = keys
		self._fbo.bind()
		self._fbo.clear_buffer()
		self._fbo.release()
		self._redraw_upd_fbo(todo)
		self._redraw_remove_fbo(get_rid)
		Logger.debug(
			f"TextureStackPlane: redrawn in "
			f"{monotonic() - start_ts:,.2f} seconds"
		)

	def iter_collided_keys(self, x, y):
		hits = (
			(self._left_xs <= x)
			& (self._bot_ys <= y)
			& (y <= self._top_ys)
			& (x <= self._right_xs)
		)
		return map(itemgetter(0), filter(itemgetter(1), zip(self._keys, hits)))


class GraphPawnSpot(ImageStackProxy, Layout):
	"""The kind of ImageStack that represents a :class:`Thing` or
	:class:`Place`.

	"""

	board = ObjectProperty()
	engine = ObjectProperty()
	selected = BooleanProperty(False)
	linecolor = ListProperty()
	selected_outline_color = ListProperty([0, 1, 1, 1])
	unselected_outline_color = ListProperty([0, 0, 0, 0])
	use_boardspace = True

	def __init__(self, **kwargs):
		if "proxy" in kwargs:
			kwargs["name"] = kwargs["proxy"].name
		super().__init__(**kwargs)
		self.bind(pos=self._position)

	def on_touch_move(self, touch):
		"""If I'm being dragged, move to follow the touch."""
		if touch.grab_current is not self:
			return False
		self.center = touch.pos
		return True

	def finalize(self, initial=True):
		"""Call this after you've created all the PawnSpot you need and are ready to add them to the board."""
		if getattr(self, "_finalized", False):
			return
		if self.proxy is None or not hasattr(self.proxy, "name"):
			Clock.schedule_once(partial(self.finalize, initial=initial), 0)
			return
		if initial:
			self.name = self.proxy.name
			if "_image_paths" in self.proxy:
				try:
					self.paths = self.proxy["_image_paths"]
				except Exception as ex:
					if not (
						isinstance(ex.args[0], str)
						and ex.args[0].startswith("Unable to load image type")
					):
						raise ex
					self.paths = self.default_image_paths
			else:
				self.paths = self.proxy.setdefault(
					"_image_paths", self.default_image_paths
				)
			zeroes = [0] * len(self.paths)
			self.offxs = self.proxy.setdefault("_offxs", zeroes)
			self.offys = self.proxy.setdefault("_offys", zeroes)
			self.proxy.connect(self._trigger_pull_from_proxy)
			self.finalize_children(initial=True)
		self._push_image_paths_binding = self.fbind(
			"paths", self._trigger_push_image_paths
		)
		self._push_offxs_binding = self.fbind(
			"offxs", self._trigger_push_offxs
		)
		self._push_offys_binding = self.fbind(
			"offys", self._trigger_push_offys
		)
		self._finalized = True

	def unfinalize(self):
		self.unbind_uid("paths", self._push_image_paths_binding)
		self.unbind_uid("offxs", self._push_offxs_binding)
		self.unbind_uid("offys", self._push_offys_binding)
		self._finalized = False

	def pull_from_proxy(self, *_):
		initial = not hasattr(self, "_finalized")
		self.unfinalize()
		for key, att in [
			("_image_paths", "paths"),
			("_offxs", "offxs"),
			("_offys", "offys"),
		]:
			if key in self.proxy and self.proxy[key] != getattr(self, att):
				setattr(self, att, self.proxy[key])
		self.finalize(initial)

	def _trigger_pull_from_proxy(self, *_, **__):
		Clock.unschedule(self.pull_from_proxy)
		Clock.schedule_once(self.pull_from_proxy, 0)

	@trigger
	def _trigger_push_image_paths(self, *_):
		self.proxy["_image_paths"] = list(self.paths)

	@trigger
	def _trigger_push_offxs(self, *_):
		self.proxy["_offxs"] = list(self.offxs)

	@trigger
	def _trigger_push_offys(self, *_):
		self.proxy["_offys"] = list(self.offys)

	def on_linecolor(self, *_):
		"""If I don't yet have the instructions for drawing the selection box
		in my canvas, put them there. In any case, set the
		:class:`Color` instruction to match my current ``linecolor``.

		"""
		if hasattr(self, "color"):
			self.color.rgba = self.linecolor
			return

		def upd_box_translate(*_):
			self.box_translate.xy = self.pos

		def upd_box_points(*_):
			self.box.points = [
				0,
				0,
				self.width,
				0,
				self.width,
				self.height,
				0,
				self.height,
				0,
				0,
			]

		self.boxgrp = boxgrp = InstructionGroup()
		self.color = Color(*self.linecolor)
		self.box_translate = Translate(*self.pos)
		boxgrp.add(PushMatrix())
		boxgrp.add(self.box_translate)
		boxgrp.add(self.color)
		self.box = Line()
		upd_box_points()
		self.bind(size=upd_box_points, pos=upd_box_translate)
		boxgrp.add(self.box)
		boxgrp.add(Color(1.0, 1.0, 1.0))
		boxgrp.add(PopMatrix())

	def on_board(self, *_):
		if not (hasattr(self, "group") and hasattr(self, "boxgrp")):
			Clock.schedule_once(self.on_board, 0)
			return
		self.canvas.add(self.group)
		self.canvas.add(self.boxgrp)

	def add_widget(self, wid, index=None, canvas=None):
		if index is None:
			for index, child in enumerate(self.children, start=1):
				if wid.priority < child.priority:
					index = len(self.children) - index
					break
		super().add_widget(wid, index=index, canvas=canvas)
		self._trigger_layout()

	def do_layout(self, *_):
		# First try to lay out my children inside of me,
		# leaving at least this much space on the sides
		xpad = self.proxy.get("_xpad", self.width / 4)
		ypad = self.proxy.get("_ypad", self.height / 4)
		self.gutter = gutter = self.proxy.get("_gutter", xpad / 2)
		height = self.height - ypad
		content_height = 0
		too_tall = False
		width = self.width - xpad
		content_width = 0
		groups = defaultdict(list)
		for child in self.children:
			group = child.proxy.get("_group", "")
			groups[group].append(child)
			if child.height > height:
				height = child.height
				too_tall = True
		piles = {}
		# Arrange the groups into piles that will fit in me vertically
		for group, members in groups.items():
			members.sort(key=lambda x: x.width * x.height, reverse=True)
			high = 0
			subgroups = []
			subgroup = []
			for member in members:
				high += member.height
				if high > height:
					subgroups.append(subgroup)
					subgroup = [member]
					high = member.height
				else:
					subgroup.append(member)
			subgroups.append(subgroup)
			content_height = max(
				(content_height, sum(wid.height for wid in subgroups[0]))
			)
			content_width += sum(
				max(wid.width for wid in subgrp) for subgrp in subgroups
			)
			piles[group] = subgroups
		self.content_width = content_width + gutter * (len(piles) - 1)
		too_wide = content_width > width
		# If I'm big enough to fit all this stuff, calculate an offset that will ensure
		# it's all centered. Otherwise just offset to my top-right so the user can still
		# reach me underneath all the pawns.
		if too_wide:
			offx = self.width
		else:
			offx = self.width / 2 - content_width / 2
		if too_tall:
			offy = self.height
		else:
			offy = self.height / 2 - content_height / 2
		for pile, subgroups in sorted(piles.items()):
			for subgroup in subgroups:
				subw = subh = 0
				for member in subgroup:
					rel_y = offy + subh
					member.rel_pos = (offx, rel_y)
					x, y = self.pos
					member.pos = x + offx, y + rel_y
					subw = max((subw, member.width))
					subh += member.height
				offx += subw
			offx += gutter

	def _position(self, *_):
		x, y = self.pos
		for child in self.children:
			offx, offy = getattr(child, "rel_pos", (0, 0))
			child.pos = x + offx, y + offy

	def on_selected(self, *_):
		if self.selected:
			self.linecolor = self.selected_outline_color
		else:
			self.linecolor = self.unselected_outline_color


class Stack:
	__slots__ = ["board", "proxy", "__self__"]

	default_image_paths = ["atlas://rltiles/floor.atlas/floor-normal"]

	def __init__(self, **kwargs):
		self.board = kwargs["board"]
		self.proxy = kwargs["proxy"]

	@property
	def paths(self):
		name = self.proxy["name"]
		plane = self._stack_plane
		datum = plane.data[plane._stack_index[name]]
		return datum["textures"]

	@paths.setter
	@mainthread
	def paths(self, v):
		name = self.proxy["name"]
		plane = self._stack_plane
		datum = plane.data[plane._stack_index[name]]
		plane.unbind_uid("data", plane._redraw_bind_uid)
		datum["textures"] = v
		insts = plane._instructions[name]
		rects = insts["rectangles"]
		group = insts["group"]
		for rect in rects:
			group.remove(rect)
		rects = insts["rectangles"] = []
		wide = datum.get("width", 0)
		tall = datum.get("height", 0)
		if v is None:
			v = self.default_image_paths
		for path in v:
			if not isinstance(path, str):
				raise TypeError("paths must be strings")
			tex = Image.load(path).texture
			w, h = tex.size
			if "width" not in datum and w > wide:
				wide = w
			if "height" not in datum and h > tall:
				tall = h
			rect = Rectangle(texture=tex, pos=self.pos, size=(wide, tall))
			rects.append(rect)
			group.add(rect)
		plane._redraw_bind_uid = plane.fbind("data", plane._trigger_redraw)

	@property
	def selected(self):
		return self._stack_plane.selected == self.proxy["name"]

	@selected.setter
	@mainthread
	def selected(self, v: bool):
		stack_plane: TextureStackPlane = self._stack_plane
		name = self.proxy["name"]
		insts = stack_plane._instructions[name]
		fbo = stack_plane._fbo
		fbo.bind()
		fbo.clear_buffer()
		if v:
			stack_plane.selected = name
			if "color0" in insts:
				insts["color0"].rgba = stack_plane.color_selected
			else:
				idx = stack_plane._stack_index[name]
				left = stack_plane._left_xs[idx]
				bot = stack_plane._bot_ys[idx]
				right = stack_plane._right_xs[idx]
				top = stack_plane._top_ys[idx]
				grp = insts["group"]
				insts["color0"] = Color(rgba=stack_plane.color_selected)
				grp.add(insts["color0"])
				insts["line"] = Line(
					points=[
						left,
						bot,
						right,
						bot,
						right,
						top,
						left,
						top,
						left,
						bot,
					]
				)
				grp.add(insts["line"])
				insts["color1"] = Color(rgba=[1.0, 1.0, 1.0, 1.0])
				grp.add(insts["color1"])
		else:
			if stack_plane.selected == self.proxy["name"]:
				stack_plane.selected = None
			if "color0" in insts:
				insts["color0"].rgba = [0.0, 0.0, 0.0, 0.0]
		fbo.release()

	@property
	def pos(self):
		stack_plane = self._stack_plane
		idx = stack_plane._stack_index[self.proxy["name"]]
		return float(stack_plane._left_xs[idx]), float(
			stack_plane._bot_ys[idx]
		)

	@pos.setter
	@mainthread
	def pos(self, xy):
		x, y = xy
		stack_plane = self._stack_plane
		stack_plane.unbind_uid("data", stack_plane._redraw_bind_uid)
		name = self.proxy["name"]
		insts = stack_plane._instructions[name]
		idx = stack_plane._stack_index[name]
		left = stack_plane._left_xs[idx]
		bot = stack_plane._bot_ys[idx]
		right = stack_plane._right_xs[idx]
		top = stack_plane._top_ys[idx]
		width = right - left
		height = top - bot
		r = x + width
		t = y + height
		stack_plane._left_xs[idx] = x
		stack_plane._bot_ys[idx] = y
		stack_plane._top_ys[idx] = t
		stack_plane._right_xs[idx] = r
		stack_plane._fbo.bind()
		stack_plane._fbo.clear_buffer()
		for rect in insts["rectangles"]:
			rect: Rectangle
			rect.pos = xy
			rect.flag_update()  # undocumented. sounds right?
		if "line" in insts:
			insts["line"].points = [x, y, r, y, r, t, x, t, x, y]
		stack_plane.data[idx]["pos"] = xy
		stack_plane._redraw_bind_uid = stack_plane.fbind(
			"data", stack_plane._trigger_redraw
		)
		stack_plane._fbo.release()

	@property
	def _stack_plane(self):
		return self.board.stack_plane

	@property
	def x(self):
		stack_plane = self._stack_plane
		idx = stack_plane._stack_index[self.proxy["name"]]
		return float(stack_plane._left_xs[idx])

	@x.setter
	def x(self, x):
		self.pos = x, self.y

	@property
	def y(self):
		stack_plane = self._stack_plane
		idx = stack_plane._stack_index[self.proxy["name"]]
		return float(stack_plane._bot_ys[idx])

	@y.setter
	def y(self, y):
		self.pos = self.x, y

	@property
	def size(self):
		stack_plane = self._stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		left = stack_plane._left_xs[idx]
		bot = stack_plane._bot_ys[idx]
		right = stack_plane._right_xs[idx]
		top = stack_plane._top_ys[idx]
		return float(right - left), float(top - bot)

	@size.setter
	def size(self, wh):
		w, h = wh
		stack_plane = self._stack_plane
		stack_plane.unbind_uid("data", stack_plane._redraw_bind_uid)
		name = self.proxy["name"]
		insts = stack_plane._instructions[name]
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		y = stack_plane._bot_ys[idx]
		r = stack_plane._right_xs[idx] = x + w
		t = stack_plane._top_ys[idx] = y + h
		for rect in insts["rectangles"]:
			rect.size = wh
		if "line" in insts:
			insts["line"].points = [x, y, r, y, r, t, x, t, x, y]
		stack_plane.data[idx]["size"] = wh
		stack_plane._redraw_bind_uid = stack_plane.fbind(
			"data", stack_plane._trigger_redraw
		)

	@property
	def width(self):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		left = stack_plane._left_xs[idx]
		right = stack_plane._right_xs[idx]
		return float(right - left)

	@width.setter
	def width(self, w):
		self.size = self.height, w

	@property
	def height(self):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		top = stack_plane._top_ys[idx]
		bot = stack_plane._bot_ys[idx]
		return float(top - bot)

	@height.setter
	def height(self, h):
		self.size = self.width, h

	@property
	def center(self):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		y = stack_plane._bot_ys[idx]
		r = stack_plane._right_xs[idx]
		t = stack_plane._top_ys[idx]
		w = r - x
		h = t - y
		return float(x + w / 2), float(y + h / 2)

	@center.setter
	def center(self, c):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		y = stack_plane._bot_ys[idx]
		r = stack_plane._right_xs[idx]
		t = stack_plane._top_ys[idx]
		w = r - x
		h = t - y
		self.pos = c[0] - w / 2, c[1] - h / 2

	@property
	def center_x(self):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		r = stack_plane._right_xs[idx]
		w = r - x
		return float(x + w / 2)

	@center_x.setter
	def center_x(self, cx):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		r = stack_plane._right_xs[idx]
		w = r - x
		self.pos = cx - w / 2, self.y

	@property
	def center_y(self):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		y = stack_plane._bot_ys[idx]
		t = stack_plane._top_ys[idx]
		h = t - y
		return float(y + h / 2)

	@center_y.setter
	def center_y(self, cy):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		y = stack_plane._bot_ys[idx]
		t = stack_plane._top_ys[idx]
		h = t - y
		self.pos = self.x, cy - h / 2

	@property
	def top(self):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		return float(stack_plane._top_ys[idx])

	@top.setter
	def top(self, t):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		y = stack_plane._bot_ys[idx]
		stack_plane._top_ys[idx] = t
		h = t - y
		stack_plane._bot_ys[idx] = t - h
		self.pos = self.x, t - h

	@property
	def right(self):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		return float(stack_plane._right_xs[idx])

	@right.setter
	def right(self, r):
		stack_plane = self.board.stack_plane
		name = self.proxy["name"]
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		stack_plane._right_xs[idx] = r
		w = r - x
		x = stack_plane._left_xs[idx] = r - w
		self.pos = x, self.y

	@property
	def name(self):
		return self.proxy["name"]

	def collide_point(self, x, y):
		pos = self.pos
		if x < pos[0] or y < pos[1]:
			return False
		w, h = self.size
		return x < pos[0] + w and y < pos[1] + h



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.clock import mainthread
from kivy.properties import NumericProperty, ObjectProperty, StringProperty
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.recycleview import RecycleView
from kivy.graphics import Color, Line

from .util import load_string_once


class RuleStepper(RecycleView):
	name = StringProperty()

	def from_rules_handled_turn(self, rules_handled_turn):
		data = [
			{
				"widget": "RuleStepperRuleButton",
				"name": "start",
				"end_tick": 0,
				"height": 40,
			}
		]
		prev_tick = 0
		for rbtyp, rules in sorted(
			rules_handled_turn.items(),
			key=lambda kv: min(kv[1].keys() or [-1]),
		):
			if not rules:
				continue
			last_entity = None
			last_rulebook = None
			typed = False
			for tick, (entity, rulebook, rule) in sorted(rules.items()):
				if tick == prev_tick:
					continue
				if not typed:
					data.append({"widget": "RulebookTypeLabel", "name": rbtyp})
					typed = True
				rulebook_per_entity = rbtyp in {"thing", "place", "portal"}
				if not rulebook_per_entity:
					if rulebook != last_rulebook:
						last_rulebook = rulebook
						data.append(
							{"widget": "RulebookLabel", "name": rulebook}
						)
				if entity != last_entity:
					last_entity = entity
					data.append({"widget": "EntityLabel", "name": entity})
				if rulebook_per_entity:
					if rulebook != last_rulebook:
						rulebook = last_rulebook
						data.append(
							{"widget": "RulebookLabel", "name": rulebook}
						)
				data.append(
					{
						"widget": "RuleStepperRuleButton",
						"name": rule,
						"start_tick": prev_tick,
						"end_tick": tick,
						"height": 40,
					}
				)
				prev_tick = tick
		self.data = data


class RuleStepperRuleButton(Button):
	name = StringProperty()
	start_tick = NumericProperty()
	end_tick = NumericProperty()
	tick = NumericProperty()
	set_tick = ObjectProperty()

	def __init__(self, **kwargs):
		super(RuleStepperRuleButton, self).__init__(**kwargs)
		self.bind(pos=self.upd_line, size=self.upd_line, tick=self.upd_line)

	def on_release(self, *args):
		self.set_tick(self.end_tick)
		self.tick = self.end_tick

	@mainthread
	def upd_line(self, *_):
		if hasattr(self, "color_inst"):
			if self.tick == self.end_tick:
				self.color_inst.rgba = [1, 0, 0, 1]
				self.line.points = [self.x, self.y, self.right, self.y]
			else:
				self.color_inst.rgba = [0, 0, 0, 0]
		else:
			with self.canvas:
				self.color_inst = Color(
					rgba=(
						[1, 0, 0, 1]
						if self.tick in (self.start_tick, self.end_tick)
						else [0, 0, 0, 0]
					)
				)
				self.line = Line(
					points=[self.x, self.top, self.right, self.top]
				)


class EntityLabel(Label):
	name = ObjectProperty()


class RulebookLabel(Label):
	name = ObjectProperty()  # rulebooks may have tuples for names


class RulebookTypeLabel(Label):
	name = StringProperty()


load_string_once("""
#:import ScrollEffect kivy.effects.scroll.ScrollEffect
<RuleStepper>:
	key_viewclass: 'widget'
	effect_cls: ScrollEffect
	RecycleGridLayout:
		cols: 1
		size_hint_y: None
		default_size_hint: 1, None
		default_height: 20
		height: self.minimum_height
<RuleStepperRuleButton>:
	text: '\\n'.join((self.name, str(self.end_tick)))
	font_size: 14
	text_size: self.width, None
	halign: 'center'
	tick: app.tick
	set_tick: app.time_travel_to_tick
<EntityLabel>:
	multiline: True
	text: str(self.name)
	text_size: self.width, None
	size: self.texture_size
	font_size: 14
	padding_x: 8
<RulebookLabel>:
	text: str(self.name)
	text_size: self.width, None
	size: self.texture_size
	font_size: 14
	bold: True
	padding_x: 4
<RulebookTypeLabel>:
	text: self.name
	text_size: self.width, None
	font_size: 16
	bold: True
	size: self.texture_size
""")



from tempfile import mkdtemp
from multiprocessing import freeze_support
from inspect import getsource

from kivy.clock import Clock
from kivy.lang.builder import Builder
from kivy.properties import BooleanProperty, NumericProperty

import networkx as nx
from networkx import grid_2d_graph

from ELiDE.game import GameApp, GameScreen, GridBoard


def make_grid() -> nx.Graph:



def game_start(engine: "LiSE.Engine") -> None:




from tempfile import mkdtemp
from multiprocessing import freeze_support
from inspect import getsource

from kivy import Logger
from kivy.clock import Clock
from kivy.lang.builder import Builder
from kivy.properties import BooleanProperty, NumericProperty

from ELiDE.game import GameApp, GameScreen, GridBoard


def remove_prefix(s: str, prefix: str):
	"""py3.8 compat"""
	if s.startswith(prefix):
		return s[len(prefix) :]
	return s


def game_start(engine) -> None:
	from random import randint, shuffle
	import networkx as nx

	engine.eternal["nonusage-limit"] = 100
	wide = engine.eternal.setdefault("max-pxcor", 36)
	high = engine.eternal.setdefault("max-pycor", 37)
	initworld = nx.grid_2d_graph(wide, high)
	# world wraps vertically
	for x in range(wide):
		initworld.add_edge((x, high - 1), (x, 0))
	# world wraps horizontally
	for y in range(high):
		initworld.add_edge((wide - 1, y), (0, y))

	locs = list(initworld.nodes.keys())
	shuffle(locs)
	peeps = {}

	for turtle in range(engine.eternal.setdefault("people", 60)):
		initworld.add_node(
			"turtle" + str(turtle),
			awareness=0,
			facing=randint(0, 3),
			location=locs.pop(),
			_image_paths=[
				"atlas://rltiles/base/unseen",
				"atlas://rltiles/body/robe_black",
			],
		)
		peeps["turtle" + str(turtle)] = True

	centers = {}

	for center in range(engine.eternal.setdefault("centers", 20)):
		initworld.add_node(
			"center" + str(center),
			location=locs.pop(),
			nonusage=0,
			_image_paths=["atlas://rltiles/dungeon/dngn_altar_xom"],
		)
		centers["center" + str(center)] = True

	phys = engine.new_character("physical", initworld)
	peep = engine.new_character("people", units={"physical": peeps})
	lit = engine.new_character("literature", units={"physical": centers})

	@peep.unit.rule(always=True)
	def wander(person):
		x, y = person.location.name
		if person["facing"] == 0:
			y += 1
		elif person["facing"] == 1:
			x += 1
		elif person["facing"] == 2:
			y -= 1
		elif person["facing"] == 3:
			x -= 1
		x %= person.engine.eternal["max-pxcor"]
		y %= person.engine.eternal["max-pycor"]
		person["location"] = (x, y)
		person["facing"] = (
			person["facing"]
			+ person.engine.randint(0, 1)
			- person.engine.randint(0, 1)
		)

	@engine.function
	def has_literature(person):
		for contained in person.location.contents():
			if contained.name.startswith("flyer") or contained.name.startswith(
				"center"
			):
				return True
		return False

	@peep.unit.rule
	def learn(person):
		person["awareness"] = (person["awareness"] + 1) % 15
		if person["awareness"] < 5:
			image_paths = [
				"atlas://rltiles/base/unseen",
				"atlas://rltiles/body/robe_black",
			]
		elif person["awareness"] < 10:
			image_paths = [
				"atlas://rltiles/base/unseen",
				"atlas://rltiles/body/robe_green",
			]
		elif person["awareness"] < 15:
			image_paths = [
				"atlas://rltiles/base/unseen",
				"atlas://rltiles/body/robe_white_green",
			]
		else:
			image_paths = [
				"atlas://rltiles/base/unseen",
				"atlas://rltiles/body/robe_green_gold",
			]
		if person["_image_paths"] != image_paths:
			person["_image_paths"] = image_paths
		person["last_learned"] = person.engine.turn
		for contained in person.location.contents():
			if contained.name.startswith("flyer") or contained.name.startswith(
				"center"
			):
				contained["last_read"] = person.engine.turn

	# this would be more pleasant as something like a partial
	@learn.trigger
	def literature_here(person):
		return person.engine.function.has_literature(person)

	@peep.unit.rule
	def unlearn(person):
		person["awareness"] = max((person["awareness"] - 1, 0))

	@unlearn.trigger
	def no_literature_here(person):
		return not person.engine.function.has_literature(person)

	@peep.unit.rule
	def write(person):
		lit_ = person.engine.character["literature"]
		maxnum = 0
		for unit in lit_.units():
			if unit.name.startswith("flyer"):
				maxnum = max((int(remove_prefix(unit.name, "flyer")), maxnum))
		scroll = person.location.new_thing(
			f"flyer{maxnum:02}",
			nonusage=0,
			_image_paths=["atlas://rltiles/scroll/scroll-0"],
		)
		lit_.add_unit(scroll)

	@write.trigger
	def activist(person):
		return person["awareness"] >= 15

	@peep.unit.rule
	def preach(person):
		for that in person.location.contents():
			if that != person and that.name.startswith("turtle"):
				that["awareness"] = (that["awareness"] + 1) % 15

	@preach.trigger
	def well_informed(person):
		return person["awareness"] >= 10

	@lit.unit.rule
	def disappear(ctr):
		ctr.delete()

	@disappear.trigger
	def unused(ctr):
		return (
			ctr.get("last_read", ctr.engine.turn) - ctr.engine.turn
			> ctr.engine.eternal["nonusage-limit"]
		)


class AwarenessGridBoard(GridBoard):
	def on_selection(self, *args):
		if not GameApp.get_running_app().placing_centers or not isinstance(
			self.selection, self.spot_cls
		):
			return
		prox = self.selection.proxy
		for contained in prox.contents():
			if contained.name.startswith("center"):
				return
		name = f"""center{max(
			int(remove_prefix(name,'center')) for name in self.character.thing.keys()
			if isinstance(name, str) and name.startswith('center')) + 1}"""
		prox.add_thing(
			name,
			nonusage=0,
			_image_paths=["atlas://rltiles/dungeon/dngn_altar_xom"],
		)


class MainGame(GameScreen):
	def on_parent(self, *args):
		if "game" not in self.ids or "physical" not in self.engine.character:
			if "physical" not in self.engine.character:
				Logger.debug("MainGame.on_parent: no physical character")
			Clock.schedule_once(self.on_parent, 0)
			return
		self.set_up()
		self.ids.game.board = AwarenessGridBoard(
			character=self.engine.character["physical"]
		)
		AwarenessApp.get_running_app().bind(turn=self._get_turn)

	def _get_turn(self, *args):
		app = AwarenessApp.get_running_app()
		if app.turn > app.end_turn:
			app.end_turn = app.turn
		self.ids.timeslider.value = app.turn

	def set_up(self):
		"""Regenerate the whole map"""
		self.engine.turn = 0
		if hasattr(self, "ran_once"):
			self.engine.eternal["people"] = int(self.ids.people.value)
			self.engine.eternal["centers"] = int(self.ids.centers.value)
			self.engine.eternal["nonusage-limit"] = int(
				self.ids.nonusage.value
			)
		else:
			self.pull_values()
		self._push_character()

	def pull_values(self, *_):
		app = GameApp.get_running_app()
		if (
			"people" not in app.engine.eternal
			or "centers" not in app.engine.eternal
			or "nonusage-limit" not in app.engine.eternal
		):
			Logger.debug("eternals not set")
			Clock.schedule_once(self.pull_values, 0)
			return
		app = GameApp.get_running_app()
		self.ids.people.value = app.engine.eternal["people"]
		self.ids.centers.value = app.engine.eternal["centers"]
		self.ids.nonusage.value = app.engine.eternal["nonusage-limit"]
		self.ran_once = True

	def _push_character(self, *args):
		board = self.ids.game.board
		if not board:
			Clock.schedule_once(self._push_character, 0)
			return
		board.character.thing.disconnect(board.update_from_thing)
		phys = AwarenessApp.get_running_app().engine.character["physical"]
		board.character = phys
		board.update()
		phys.thing.connect(board.update_from_thing)


class AwarenessApp(GameApp):
	play = BooleanProperty(False)
	placing_centers = BooleanProperty(False)
	end_turn = NumericProperty(0)
	inspector = True

	def on_play(self, *args):
		if self.play:
			Clock.schedule_interval(self.next_turn, self.turn_length)
			self._scheduled_next_turn = True
		elif hasattr(self, "_scheduled_next_turn"):
			Clock.unschedule(self.next_turn)
			del self._scheduled_next_turn

	def on_turn(self, *args):
		turn = int(self.turn)
		if turn != self.engine.turn:
			self.engine.turn = turn

	def on_start(self):
		if not hasattr(self, "_initialized"):
			self._initialized = True
			self.engine.game_init()


kv = """
# kv_start
<ScreenManager>:
	MainGame:
		name: 'play'
<MainGame>:
	BoxLayout:
		orientation: 'vertical'
		BoxLayout:
			orientation: 'horizontal'
			BoxLayout:
				id: sidebar
				orientation: 'vertical'
				width: 333
				size_hint_x: None
				BoxLayout:
					id: controls
					orientation: 'vertical'
					size_hint_y: 0.3
					BoxLayout:
						id: toprow
						Button:
							id: setup
							text: 'setup'
							on_release: root.set_up()
						Slider:
							id: people
							min: 0
							max: 300
							step: 1
							Label:
								text: 'people'
								x: people.x
								y: people.y
								size: self.texture_size
								size_hint: None, None
							Label:
								text: str(int(people.value))
								x: people.right - self.texture_size[0]
								y: people.y
								size: self.texture_size
								size_hint: None, None
						ToggleButton:
							text: 'place centers'
							on_release: app.placing_centers = self.state == 'down'
					BoxLayout:
						id: midrow	
						ToggleButton:
							id: play
							text: 'Go'
							on_state: app.play = self.state == 'down'
						Slider:
							id: centers
							min: 0
							max: 100
							step: 1
							Label:
								text: 'centers'
								size: self.texture_size
								pos: self.parent.pos
							Label:
								text: str(int(self.parent.value))
								size: self.texture_size
								center_x: self.parent.center_x
								y: self.parent.y
						Widget:
							id: filler0
					Slider:
						id: nonusage
						min: 0
						max: 500
						step: 1
						Label:
							text: 'non-usage limit'
							size: self.texture_size
							pos: nonusage.pos
						Label:
							text: '{:d} ticks'.format(int(nonusage.value))
							size: self.texture_size
							x: nonusage.right - self.texture_size[0]
							y: nonusage.y
				Widget:
					id: filler
			Widget:
				id: gamebox
				size_hint_x: 0.7
				GridBoardView:
					id: game		
					pos: gamebox.pos
					size: gamebox.size
		Slider:
			id: timeslider
			x: 0
			y: 0
			height: 50
			size_hint_y: None
			min: 0
			max: app.end_turn
			step: 1
			on_value: app.turn = self.value
# kv_end
"""

Builder.load_string(kv)

if __name__ == "__main__":
	freeze_support()
	d = mkdtemp()
	with open(d + "/game_start.py", "w", encoding="utf-8") as outf:
		outf.write(getsource(game_start))
	AwarenessApp(prefix=d).run()
	print("Files are in " + d)



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial

from kivy.app import App
from kivy.clock import Clock
from kivy.uix.screenmanager import Screen
from kivy.uix.recycleview import RecycleView
from kivy.properties import (
	ListProperty,
	ObjectProperty,
	StringProperty,
)

from .util import SelectableRecycleBoxLayout, load_string_once

# TODO: Visual preview
# TODO: Background image chooser


class CharactersRecycleBoxLayout(SelectableRecycleBoxLayout):
	character_name = StringProperty()

	def apply_selection(self, index, view, is_selected):
		super().apply_selection(index, view, is_selected)
		if is_selected:
			self.character_name = view.text


class CharactersView(RecycleView):
	character_name = StringProperty()

	def __init__(self, **kwargs):
		self.i2name = {}
		self.name2i = {}
		super().__init__(**kwargs)


class CharactersScreen(Screen):
	toggle = ObjectProperty()
	charsview = ObjectProperty()
	character_name = StringProperty()
	wallpaper_path = StringProperty()
	names = ListProperty()
	new_board = ObjectProperty()
	push_character_name = ObjectProperty()

	@property
	def engine(self):
		return App.get_running_app().engine

	def new_character(self, name, *_):
		self.engine.add_character(name)
		self.ids.newname.text = ""
		i = len(self.charsview.data)
		self.charsview.i2name[i] = name
		self.charsview.name2i[name] = i
		self.charsview.data.append({"index": i, "text": name})
		self.names.append(name)
		self.new_board(name)
		self.push_character_name(name)

	def _trigger_new_character(self, name):
		part = partial(self.new_character, name)
		if hasattr(self, "_scheduled_new_character"):
			Clock.unschedule(self._scheduled_new_character)
		self._scheduled_new_character = Clock.schedule_once(part)

	def _munge_names(self, names):
		for i, name in enumerate(names):
			self.charsview.i2name[i] = name
			self.charsview.name2i[name] = i
			yield {"index": i, "text": name}

	def on_names(self, *_):
		app = App.get_running_app()
		if not app.character or not self.charsview:
			Clock.schedule_once(self.on_names, 0)
			return
		self.charsview.data = list(self._munge_names(self.names))
		charname = app.character.name
		for i, name in enumerate(self.names):
			if name == charname:
				self.charsview.children[0].select_node(i)
				return

	def on_charsview(self, *_):
		if not self.push_character_name:
			Clock.schedule_once(self.on_charsview, 0)
			return
		self.charsview.bind(character_name=self.setter("character_name"))
		self.bind(character_name=self.push_character_name)


load_string_once("""
#: import resource_find kivy.resources.resource_find
<CharactersView>:
	viewclass: 'RecycleToggleButton'
	character_name: boxl.character_name
	CharactersRecycleBoxLayout:
		id: boxl
		multiselect: False
		default_size: None, dp(56)
		default_size_hint: 1, None
		size_hint_y: None
		height: self.minimum_height
		orientation: 'vertical'
<CharactersScreen>:
	name: 'chars'
	charsview: charsview
	BoxLayout:
		id: chars
		orientation: 'vertical'
		CharactersView:
			id: charsview
			size_hint_y: 0.8
			character_name: root.character_name
		TextInput:
			id: newname
			size_hint_y: 0.1
			hint_text: 'New character name'
			write_tab: False
			multiline: False
		Button:
			text: '+'
			on_release: root._trigger_new_character(newname.text)
			size_hint_y: 0.05
		Button:
			text: 'Close'
			on_release: root.toggle()
			size_hint_y: 0.05
""")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Generic dialog boxes and menus, for in front of a Board"""

from functools import partial
from kivy.properties import (
	DictProperty,
	ListProperty,
	ObjectProperty,
	StringProperty,
	NumericProperty,
	VariableListProperty,
)
from kivy.core.text import DEFAULT_FONT
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.widget import Widget
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.scrollview import ScrollView
from kivy.clock import Clock, mainthread
from kivy.logger import Logger

from .util import load_string_once


class Box(Widget):
	padding = VariableListProperty(6)
	border = VariableListProperty(4)
	font_size = StringProperty("15sp")
	font_name = StringProperty(DEFAULT_FONT)
	background = StringProperty()
	background_color = VariableListProperty([1, 1, 1, 1])
	foreground_color = VariableListProperty([0, 0, 0, 1])


class ScrollableLabel(ScrollView):
	font_size = StringProperty("15sp")
	font_name = StringProperty(DEFAULT_FONT)
	color = VariableListProperty([0, 0, 0, 1])
	line_spacing = NumericProperty(0)
	text = StringProperty()


class MessageBox(Box):
	"""Looks like a TextInput but doesn't accept any input.

	Does support styled text with BBcode.

	"""

	line_spacing = NumericProperty(0)
	text = StringProperty()


class DialogMenu(Box):
	"""Some buttons that make the game do things.

	Set ``options`` to a list of pairs of ``(text, function)`` and the
	menu will be populated with buttons that say ``text`` that call
	``function`` when pressed.

	"""

	options = ListProperty()
	"""List of pairs of (button_text, callable)"""

	def _set_sv_size(self, *_):
		self._sv.width = self.width - self.padding[0] - self.padding[2]
		self._sv.height = self.height - self.padding[1] - self.padding[3]

	def _set_sv_pos(self, *_):
		self._sv.x = self.x + self.padding[0]
		self._sv.y = self.y + self.padding[3]

	@mainthread
	def on_options(self, *_):
		if not hasattr(self, "_sv"):
			self._sv = ScrollView(size=self.size, pos=self.pos)
			self.bind(size=self._set_sv_size, pos=self._set_sv_pos)
			layout = BoxLayout(orientation="vertical")
			self._sv.add_widget(layout)
			self.add_widget(self._sv)
		else:
			layout = self._sv.children[0]
			layout.clear_widgets()
		for txt, part in self.options:
			if not callable(part):
				raise TypeError("Menu options must be callable")
			butn = Button(
				text=txt,
				on_release=part,
				font_name=self.font_name,
				font_size=self.font_size,
				valign="center",
				halign="center",
			)
			butn.bind(size=butn.setter("text_size"))
			layout.add_widget(butn)


class Dialog(BoxLayout):
	"""MessageBox with a DialogMenu beneath it.

	Set the properties ``message_kwargs`` and ``menu_kwargs``,
	respectively, to control them -- but you probably want
	to do that by returning a pair of dicts from an action
	in LiSE.

	"""

	message_kwargs = DictProperty({})
	menu_kwargs = DictProperty({})

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(
			message_kwargs=self._propagate_msg_kwargs,
			menu_kwargs=self._propagate_menu_kwargs,
		)
		self._propagate_msg_kwargs()
		self._propagate_menu_kwargs()

	def _propagate_msg_kwargs(self, *_):
		if "msg" not in self.ids:
			Clock.schedule_once(self._propagate_msg_kwargs, 0)
			return
		kw = dict(self.message_kwargs)
		kw.setdefault(
			"background", "atlas://data/images/defaulttheme/textinput"
		)
		for k, v in kw.items():
			setattr(self.ids.msg, k, v)

	def _propagate_menu_kwargs(self, *_):
		if "menu" not in self.ids:
			Clock.schedule_once(self._propagate_menu_kwargs, 0)
			return
		kw = dict(self.menu_kwargs)
		kw.setdefault(
			"background",
			"atlas://data/images/defaulttheme/vkeyboard_background",
		)
		for k, v in kw.items():
			setattr(self.ids.menu, k, v)


class DialogLayout(FloatLayout):
	"""A layout, normally empty, that can generate dialogs

	To make dialogs, set my ``todo`` property to a list. It may contain:

	* Strings, which will be displayed with an "OK" button to dismiss them
	* Lists of pairs of strings and callables, which generate buttons with the
	  string on them that, when clicked, call the callable
	* Lists of pairs of dictionaries, which are interpreted as keyword
	  arguments to :class:`MessageBox` and :class:`DialogMenu`

	In place of a callable you can use the name of a function in my
	``usermod``, a Python module given by name. I'll import it when I need it.

	Needs to be instantiated with a LiSE ``engine`` -- probably an
	:class:`EngineProxy`.

	"""

	dialog = ObjectProperty()
	engine = ObjectProperty()
	todo = ListProperty()
	usermod = StringProperty("user")
	userpkg = StringProperty(None, allownone=True)

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.dialog = Dialog()

	def on_engine(self, *_):
		todo = self.engine.universal.get("last_result")
		if isinstance(todo, list):
			self.todo = todo
		else:
			self.todo = []
		self.idx = self.engine.universal.get("last_result_idx", 0)
		self.engine.universal.connect(self._pull)
		if self.todo:
			self.advance_dialog()

	def _pull(self, *_, key, value):
		if key == "last_result":
			self.todo = value if value and isinstance(value, list) else []
		elif key == "last_result_idx":
			self.idx = value if value and isinstance(value, int) else 0

	def on_idx(self, *_):
		lidx = self.engine.universal.get("last_result_idx")
		if lidx is not None and lidx != self.idx:
			self.engine.universal["last_result_idx"] = self.idx
		Logger.debug(f"DialogLayout.idx = {self.idx}")

	def advance_dialog(self, after_ok=None, *args):
		"""Try to display the next dialog described in my ``todo``.

		:param after_ok: An optional callable. Will be called after the user clicks
		a dialog option--as well as after any game-specific code for that option
		has run.

		"""
		self.clear_widgets()
		try:
			Logger.debug(f"About to update dialog: {self.todo[0]}")
			self._update_dialog(self.todo.pop(0), after_ok)
		except IndexError:
			if after_ok is not None:
				after_ok()

	@mainthread
	def _update_dialog(self, diargs, after_ok, **kwargs):
		if diargs is None:
			Logger.debug("DialogLayout: null dialog")
			return
		if isinstance(diargs, tuple) and diargs[0] == "stop":
			if len(diargs) == 1:
				Logger.debug("DialogLayout: null dialog")
				return
			diargs = diargs[1]
		dia = self.dialog
		# Simple text dialogs just tell the player something and let them click OK
		if isinstance(diargs, str):
			dia.message_kwargs = {"text": diargs}
			dia.menu_kwargs = {
				"options": [("OK", partial(self.ok, cb2=after_ok))]
			}
		# List dialogs are for when you need the player to make a choice and don't care much
		# about presentation
		elif isinstance(diargs, list):
			dia.message_kwargs = {"text": "Select from the following:"}
			dia.menu_kwargs = {
				"options": list(
					map(partial(self._munge_menu_option, after_ok), diargs)
				)
			}
		# For real control of the dialog, you need a pair of dicts --
		# the 0th describes the message shown to the player, the 1th
		# describes the menu below
		elif isinstance(diargs, tuple):
			if len(diargs) != 2:
				# TODO more informative error
				raise TypeError("Need a tuple of length 2")
			msgkwargs, mnukwargs = diargs
			if isinstance(msgkwargs, dict):
				dia.message_kwargs = msgkwargs
			elif isinstance(msgkwargs, str):
				dia.message_kwargs["text"] = msgkwargs
			else:
				raise TypeError("Message must be dict or str")
			if isinstance(mnukwargs, dict):
				mnukwargs["options"] = list(
					map(
						partial(self._munge_menu_option, after_ok),
						mnukwargs["options"],
					)
				)
				dia.menu_kwargs = mnukwargs
			elif isinstance(mnukwargs, (list, tuple)):
				dia.menu_kwargs["options"] = list(
					map(partial(self._munge_menu_option, after_ok), mnukwargs)
				)
			else:
				raise TypeError("Menu must be dict or list")
		else:
			raise TypeError(
				"Don't know how to turn {} into a dialog".format(type(diargs))
			)
		self.add_widget(dia)

	def ok(self, *_, cb=None, cb2=None):
		"""Clear dialog widgets, call ``cb``s if provided, and advance the dialog queue"""
		self.clear_widgets()
		if cb:
			cb()
		if cb2:
			cb2()

	def _lookup_func(self, funcname):
		from importlib import import_module

		if not hasattr(self, "_usermod"):
			self._usermod = import_module(self.usermod, self.userpkg)
		return getattr(self.usermod, funcname)

	def _munge_menu_option(self, after_ok, option):
		if not isinstance(option, tuple):
			raise TypeError
		name, func = option
		if func is None:
			return name, partial(self.ok, cb2=after_ok)
		if callable(func):
			return name, partial(self.ok, cb=func, cb2=after_ok)
		if isinstance(func, tuple):
			fun = func[0]
			if isinstance(fun, str):
				fun = self._lookup_func(fun)
			args = func[1]
			if len(func) == 3:
				kwargs = func[2]
				func = partial(fun, *args, **kwargs)
			else:
				func = partial(fun, *args)
		if isinstance(func, str):
			func = self._lookup_func(func)
		return name, partial(self.ok, cb=func, cb2=after_ok)


load_string_once("""
<Box>:
	canvas.before:
		Color:
			rgba: self.background_color
		BorderImage:
			border: self.border
			pos: self.pos
			size: self.size
			source: self.background
		Color:
			rgba: 1, 1, 1, 1
<ScrollableLabel>:
	Label:
		size_hint_y: None
		height: self.texture_size[1]
		text_size: self.width, None
		text: root.text
		color: root.color
<MessageBox>:
	ScrollableLabel:
		x: root.x + root.padding[0]
		y: root.y + root.padding[3]
		width: root.width - root.padding[2]
		height: root.height - root.padding[1]
		text: root.text
		color: root.foreground_color
<Dialog>:
	orientation: 'vertical'
	pos_hint: {'x': 0, 'y': 0}
	size_hint: 1, 0.3
	MessageBox:
		id: msg
	DialogMenu:
		id: menu
""")
